// Summary: Theta neuron class definition
// Author: Sam McKennoch
// Date: 2007-2009

/*! \class thetaNeuron
 *  \brief Class to encapsulate the behavior of a single theta neuron
 *
 * This class, in addition to setting, getting and displaying neuron properties, 
 * includes functionality to simulate a Theta Neuron.  There are two simulation methods
 * (methods by which to observe the timing of output spikes which result from the interaction of 
 * the neuron dynamic properties and input spikes) 
 * available, Numerical and EventDriven.  These methods produce nearly identical results, however EventDriven
 * is more efficient, while Numerical is included for clarity and comparison.  
 * There will be at most one output spike after the last input spike, depending on the neuron
 * phase and the baseline current.
 *  \author Sam McKennoch
 *  \version 9.0
 *  \date    2007-2009
 *  \warning Only Minimal Error Handling is Currently Implemented
 *
 *  \example thetaNeuronExample.cpp
 *  This is an example of how to use the thetaNeuron class.
 *  This example does the follows:
 *		- Construct Theta Neuron Object with Randomized Weights
 *		- Displays Neuron Properties
 *		- Displays Neuron Phase (State)
 *		- Simulates Theta Neuron Numerically
 *		- Simulates Theta Neuron with Event-Driven Methodology
 *		- Finds Mean-Squared in Spike Times Between the Different Methods
 *		- Stacks Two Neurons Sequentially
 *		- Destructs Theta Neuron Object
 */

#ifndef THETANEURON_HEADER
#define THETANEURON_HEADER

//standard libraries
#include "stdafx.h"
#include <fstream>
#include <string>
#include <iostream>
#include <cmath>
#include <ctime>

using namespace std;

//Constant Defaults
#define IO -0.001
#define ALPHA 1
#define OFFSET 0.0001
#define TIMESTEP 0.001
#define EPS 1e-10
#define PI 3.14159265
#define MAX_NUM_OUTPUT_SPIKES 5


class thetaNeuron  {

//----------------------------------------------------------------------------------------------------------------
//public methods
//----------------------------------------------------------------------------------------------------------------
public:

	/** 
	* Simulation Method Enum.
	*
	* This Enumeration contains the two possible methods used for simulating theta neurons in this implementation, Numerical and EventDriven.
	*/
	enum Method
	{
		Numerical, /**< enum value Numerical. */
		EventDriven /**< enum value EventDriven. */
	};

	/**
	* Theta Neuron Constructor.
	*
	* The theta neuron constructor takes three inputs, the number of neuron inputs, the initial weight and the baseline current.  
	* 
	* In this implementation, new_io must be negative, with typical values of -0.0001 < io < -0.1.  If an attempt is made to assign io 
	* to a non-negative value or if io is not assigned at all, then io reverts to the default value of -0.001.  
	*
	* If weight_ini is not specified or if it is specified as > 100, then the weights are randomized with 
	* a uniform distribution between -1 and 1.
	*
	* If new_num_inputs is not specified, then the default value of 3 input neurons is used.
	*
	* Other neuron properties are also assigned here and may be subsequently change using the get/set functions.  
	* Specifically, verbose is set to 1 meaning additional information will be displayed to the console during simulation, 
	* and method is set to Numerical, meaning numerical intergation will be used by default in the simulation.
	*
	* @param new_num_inputs The number of input neurons desired.
	* @param weight_ini The value to initialize all weights to.
	* @param new_io The baseline current. 
	*/
	thetaNeuron(int new_num_inputs = 3.0, double weight_ini = 1e6, double new_io = IO){
		srand((unsigned)time(0));
		weights=new double[new_num_inputs];
		for(int k=0;k<new_num_inputs;k++){
			if (weight_ini>100)
				weights[k] = (2.0*rand()/(RAND_MAX + 1.0))-1.0;
			else
				weights[k] = weight_ini;
		} 
		num_inputs=new_num_inputs;
		io = new_io;
		if (io>=0) io=IO;
		beta=sqrt(fabs(ALPHA*io));
		positive_fixed_point=acos((1+ALPHA*io)/(1-ALPHA*io));
		phase=positive_fixed_point+OFFSET;
		verbose=1;                                       
		method=Numerical;
	}

	/**
	* Theta Neuron Destructor.
	*
	* The Destructor simply deallocates the memory used for the neuron weights.
	*/
	~thetaNeuron(){ 
		delete [] weights; 
		weights = NULL;
	}


	//Get and Set Methods

	/**
	* Public Method to Get the Neuron Phase (State)
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return The neuron phase
	*/
	double getPhase(void) {return phase;}

	/**
	* Public Method to Get the Neuron Simulation Method
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return The neuron simulation method
	*/	
	Method getMethod(void) {return method;}

	/**
	* Public Method to Set the Neuron Simulation Method
	* @param new_method An enumerator of type Method representing the new simulation method
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return void
	*/
	void setMethod(Method new_method) {method=new_method;}

	/**
	* Public Method to Set the Verbosity of the Theta Neuron Simulation
	* @param new_verbose A boolean flag representing whether verbose comments should be printed to the console window during simulation
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return void
	*/
	void setVerbose(bool new_verbose) {verbose=new_verbose;}

	/**
	* Public Method to Get the Weight at input neuron index
	* @param index Index into the neuron's inputs from which to get the weight value; must not be greater than the number of inputs or -1 will be returned
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return The weight at input neuronindex
	*/	
	double getWeight(int index) {
		if (index<num_inputs) {
			return weights[index];
		}
		else return -1;
	}

	/**
	* Public Method to Set the Weight at input neuron index to new_weight
	* @param new_weight New weight value
	* @param index Index into neuron's inputs at which to set the new weight; must not be greater than the number of inputs or weight will not be set
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return void
	*/
	void setWeight(double new_weight, int index) {
		if (index<num_inputs) {
			weights[index]=new_weight;
		}
	}

	/**
	* Public Method to Get the Baseline Current, Io
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return The baseline current, Io
	*/	
	double getIo(void) {return io;}

	/**
	* Public Method to Set the Baseline Current, Io
	* @param new_io New baseline current value, must be negative or baseline current will not be set
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return void
	*/	
	void setIo(double new_io) {
		if (new_io<0) {
			io=new_io;
			beta=sqrt(fabs(ALPHA*io));
			positive_fixed_point=acos((1+ALPHA*io)/(1-ALPHA*io));
			phase=positive_fixed_point+OFFSET;
		}
	}	

	/**
	* Public Method to Display Theta Neuron Properties
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return void
	*/
	void displayThetaNeuron(void);

	/**
	* Public Method to Find Output Spike Times, ts, of a maximum length max_num_ts, with input spike times in array ti
	* @param ti an array of input spike times.
	* @param ts an array that will be filled in with output spike times.
	* @param max_num_ts the maximum number of output spikes that will be returned.
	* @see thetaNeuron()
	* @see ~thetaNeuron()
	* @return The output spike times
	*/
	int runThetaNeuron(double ti[], double ts[], int max_num_ts);

//----------------------------------------------------------------------------------------------------------------
//private members
//----------------------------------------------------------------------------------------------------------------
private:      
	double io; /*!< The Baseline Current*/
	double beta; /*!< A Common Theta Neuron Property*/
	double positive_fixed_point; /*!< A Theta Phase Saddle Point*/
	double phase; /*!< The State of the Theta Neuron*/   
	double *weights; /*!< An Array of Synaptic Weights on the Theta Neuron's Inputs*/
	int num_inputs; /*!< The Number of Inputs to the Theta Neuron*/
	bool verbose; /*!< A Flag To Indicate If Detailed Information Should Be Printed to the Console During Neuron Simulation*/
	Method method; /*!< An Enumerator Indicating the Method by which to Simulate the Theta Neuron*/


	//Simple Implementation of atanh
	//Assumes -1 < x < 1
	static double atanh(double x)
	{
		return ( 0.5 * log( ( 1.0 + x ) / ( 1.0 - x ) ) );
	}

	//Function perform descending order bubble sort on array
	//The indices sorted indices are returned as well.
	//For example, if array={1.5 0.5 7}, then array will shuffled
	//into {0.5 1.5 7} and indices will be {1 0 2}
	void sortArray(double array[], int indices[], int size)
	{
		bool swap;
		double temp;
		int temp2;
		int n;
		for (n = 0; n < size; n++)
		{
			indices[n]=n;
		}
		do
		{
			swap = false;
			for (n = 0; n < (size - 1); n++)
			{
				if (array[n] > array[n + 1])
				{
					temp = array[n];
					array[n] = array[n + 1];
					array[n + 1] = temp;
					temp2 = indices[n];
					indices[n] = indices[n + 1];
					indices[n + 1] = temp2;            
					swap = true;
				}
			} 
		} while (swap);
	}

	//Template Function to Display Array Members on One Line in the Console
	//size is the size of array
	template <class ItemType>
	void showArray(ItemType array[], int size)
	{
		for (int j = 0; j < size; j++) {
			cout << array[j] << " ";
		}
		cout << endl;
	}	
};

#endif